@@ -31,7 +31,6 @@ import tempfile
 
 from gettext import gettext as _
 from threading import Thread
-from traceback import print_exc
 
 from softwareproperties.shortcuts import ShortcutException
 
@@ -53,6 +52,7 @@ LAUNCHPAD_PPA_API = 'https://launchpad.net/api/1.0/%s/+archive/%s'
 LAUNCHPAD_USER_API = 'https://launchpad.net/api/1.0/%s'
 LAUNCHPAD_USER_PPAS_API = 'https://launchpad.net/api/1.0/%s/ppas'
 LAUNCHPAD_DISTRIBUTION_API = 'https://launchpad.net/api/1.0/%s'
+LAUNCHPAD_DISTRIBUTION_SERIES_API = 'https://launchpad.net/api/1.0/%s/%s'
 # Specify to use the system default SSL store; change to a different path
 # to test with custom certificates.
 LAUNCHPAD_PPA_CERT = "/etc/ssl/certs/ca-certificates.crt"
@@ -88,9 +88,19 @@ def get_info_from_lp(lp_url):
         return _get_https_content_py3(lp_url)
 
 def get_ppa_info_from_lp(owner_name, ppa):
+    if owner_name[0] != '~':
+        owner_name = '~' + owner_name
     lp_url = LAUNCHPAD_PPA_API % (owner_name, ppa)
     return get_info_from_lp(lp_url)
 
+def series_valid_for_distro(distribution, series):
+    lp_url = LAUNCHPAD_DISTRIBUTION_SERIES_API % (distribution, series)
+    try:
+        get_info_from_lp(lp_url)
+        return True
+    except PPAException:
+        return False
+
 def get_current_series_from_lp(distribution):
     lp_url = LAUNCHPAD_DISTRIBUTION_API % distribution
     return os.path.basename(get_info_from_lp(lp_url)["current_series_link"])
@@ -121,10 +131,13 @@ def _get_https_content_pycurl(lp_url):
         curl.setopt(pycurl.URL, str(lp_url))
         curl.setopt(pycurl.HTTPHEADER, ["Accept: application/json"])
         curl.perform()
+        response = curl.getinfo(curl.RESPONSE_CODE)
         curl.close()
         json_data = callback.contents
     except pycurl.error as e:
         raise PPAException("Error reading %s: %s" % (lp_url, e), e)
+    if response != 200:
+        raise PPAException("Error reading %s: response code %i" % (lp_url, response))
     return json.loads(json_data)
 
 
@@ -166,7 +179,7 @@ class AddPPASigningKey(object):
         try:
             # double check that the signing key is a v4 fingerprint (160bit)
             if not verify_keyid_is_v4(signing_key_fingerprint):
-                print("Error: signing key fingerprint '%s' too short" % 
+                print("Error: signing key fingerprint '%s' too short" %
                     signing_key_fingerprint)
                 return False
         except TypeError:
@@ -193,6 +206,18 @@ class AddPPASigningKey(object):
             return False
         return True
 
+    def _export_armor_key(self, keyring, export_keyring, signing_key_fingerprint, keyring_dir):
+        res = subprocess.call(self.GPG_DEFAULT_OPTIONS + [
+          "--homedir", keyring_dir,
+          "--keyring", keyring,
+          "--output", export_keyring,
+          "--armor",
+          "--export", signing_key_fingerprint,
+          ])
+        if res != 0:
+            return False
+        return True
+
     def _get_fingerprints(self, keyring, keyring_dir):
         cmd = self.GPG_DEFAULT_OPTIONS + [
           "--homedir", keyring_dir,
@@ -211,7 +236,7 @@ class AddPPASigningKey(object):
     def _verify_fingerprint(self, keyring, expected_fingerprint, keyring_dir):
         got_fingerprints = self._get_fingerprints(keyring, keyring_dir)
         if len(got_fingerprints) > 1:
-            print("Got '%s' fingerprints, expected only one" % 
+            print("Got '%s' fingerprints, expected only one" %
                   len(got_fingerprints))
             return False
         got_fingerprint = got_fingerprints[0]
@@ -223,7 +248,7 @@ class AddPPASigningKey(object):
 
     def add_ppa_signing_key(self, ppa_path=None):
         """Query and add the corresponding PPA signing key.
-        
+
         The signing key fingerprint is obtained from the Launchpad PPA page,
         via a secure channel, so it can be trusted.
         """
@@ -263,12 +288,18 @@ class AddPPASigningKey(object):
             tmp_export_keyring, signing_key_fingerprint, tmp_keyring_dir):
             cleanup(tmp_keyring_dir)
             return False
+        # now export armor key, because apt-key cannot import keybox
+        tmp_export_armor_keyring = os.path.join(tmp_keyring_dir, "export-armor-keyring.gpg")
+        if not self._export_armor_key(
+            tmp_keyring, tmp_export_armor_keyring, signing_key_fingerprint, tmp_keyring_dir):
+            cleanup(tmp_keyring_dir)
+            return False
         # and add it
         trustedgpgd = apt_pkg.config.find_dir("Dir::Etc::trustedparts")
         apt_keyring = os.path.join(trustedgpgd, "%s.gpg" % (
             encode(ppa_info["reference"][1:])))
         res = subprocess.call(["apt-key", "--keyring", apt_keyring, "add",
-            tmp_keyring])
+            tmp_export_armor_keyring])
         # cleanup
         cleanup(tmp_keyring_dir)
         return (res == 0)
@@ -365,9 +396,16 @@ class PPAShortcutHandler(object):
         return self._info
 
     def expand(self, codename, distro=None):
-        if distro is not None and distro != self._info["distribution"]:
+        if (distro is not None
+                and distro != self._info["distribution"]
+                and not series_valid_for_distro(self._info["distribution"], codename)):
             # The requested PPA is for a foreign distribution.  Guess that
             # the user wants that distribution's current series.
+            # This only applies if the local distribution is not the same
+            # distribution the remote PPA is associated with AND the local
+            # codename is not equal to the PPA's series.
+            # e.g. local:Foobar/xenial and ppa:Ubuntu/xenial will use 'xenial'
+            #      local:Foobar/fluffy and ppa:Ubuntu/xenial will use '$latest'
             codename = get_current_series_from_lp(self._info["distribution"])
         debline = "deb http://ppa.launchpad.net/%s/%s/%s %s main" % (
             self._info["owner"][1:], self._info["name"],
@@ -389,11 +427,7 @@ class PPAShortcutHandler(object):
 def shortcut_handler(shortcut):
     if not shortcut.startswith("ppa:"):
         return None
-    try:
-        return PPAShortcutHandler(shortcut)
-    except ShortcutException:
-        print_exc()
-        return None
+    return PPAShortcutHandler(shortcut)
 
 
 if __name__ == "__main__":
